/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * @(#)JavaMonkey.cpp	1.12 03/09/02
 */
 
/* 
 * A sample Assistive Technology which builds a tree of all of the 
 * accessible UI objects in the JavaVM using the AccessBridge technology
 */

#include <windows.h>   // includes basic windows functionality
#include <stdio.h>
#include <commctrl.h>
#include <jni.h>
#include "monkeyResource.h"
#include "AccessBridgeCalls.h"
#include "AccessBridgeCallbacks.h"

#define BUFSIZE 4096 * 4

// -----------------------------

/**
 * A node in the Monkey tree
 */
class AccessibleNode {
	HWND baseHWND;
	HTREEITEM treeNodeParent;
	long vmID;
	AccessibleContext ac;
	AccessibleNode *parentNode;
	char accessibleName[MAX_STRING_SIZE];
	char accessibleRole[SHORT_STRING_SIZE];
public:
	AccessibleNode(long vmID, AccessibleContext context, 
				   AccessibleNode *parent, HWND hWnd, 
				   HTREEITEM parentTreeNodeItem);
	~AccessibleNode();
	void setAccessibleName(char *name);
	void setAccessibleRole(char *role);

	BOOL displayAPIWindow();	// bring up an Accessibility API detail window
};


/**
 * The main application class
 */
class JavaMonkey {
	HINSTANCE theInstance;
public:
	JavaMonkey(HINSTANCE hInstance);
	BOOL InitWindow(int windowMode);
	char *getAccessibleInfo(long vmID, AccessibleContext ac, char *buffer, int bufsize);
	void exitJavaMonkey(HWND hWnd);
	void buildAccessibilityTree();
	void addComponentNodes(long vmID, AccessibleContext context, 
						   AccessibleNode *parent, HWND hWnd,
						   HTREEITEM treeNodeParent, HWND treeWnd);
};

// -----------------------------

extern "C" {

	HWND theMonkeyWindow;
	HWND theTreeWindowControl;
	HINSTANCE theInstance;
	JavaMonkey *theMonkey;
	AccessibleNode *theSelectedNode;
	AccessibleContext theSelectedAccessibleContext;

	LRESULT CALLBACK MonkeyWindowProc(HWND hDlg, UINT message, UINT wParam, LONG lParam);
	BOOL CALLBACK EnumWndProc(HWND hWnd, LPARAM lParam);
	LRESULT CALLBACK ExplorerWindowProc (HWND hWnd, UINT message, UINT wParam, LONG lParam);

	char *getAccessibleInfo(long vmID, AccessibleContext ac, char *buffer, int bufsize);

	char theMonkeyClassName[] = "MonkeyWin";
	char theExplorerClassName[] = "ExplorerWin";
}

// -----------------------------

JavaMonkey::JavaMonkey(HINSTANCE hInstance) {
	WNDCLASSEX wcl;
	
	theInstance = hInstance;

	// define the window classes

	// Monkey Window
	wcl.cbSize = sizeof(WNDCLASSEX);

	wcl.hInstance = hInstance;
	wcl.lpszClassName = theMonkeyClassName;
	wcl.lpfnWndProc = MonkeyWindowProc;
	wcl.style = 0;

	wcl.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wcl.hIconSm = LoadIcon(NULL, IDI_APPLICATION);
	wcl.hCursor = LoadCursor(NULL, IDC_ARROW);

	wcl.lpszMenuName = "MONKEYMENU";
	wcl.cbClsExtra = 0;
	wcl.cbWndExtra = 0;

	wcl.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);

	RegisterClassEx(&wcl);

	// Explorer Window
	wcl.cbSize = sizeof(WNDCLASSEX);

	wcl.hInstance = hInstance;
	wcl.lpszClassName = theExplorerClassName;
	wcl.lpfnWndProc = ExplorerWindowProc;
	wcl.style = 0;

	wcl.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wcl.hIconSm = LoadIcon(NULL, IDI_APPLICATION);
	wcl.hCursor = LoadCursor(NULL, IDC_ARROW);

	wcl.lpszMenuName = "";
	wcl.cbClsExtra = 0;
	wcl.cbWndExtra = 0;

	wcl.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);

	RegisterClassEx(&wcl);
}

/**
 * Real(tm) MS-Windows window initialization
 *
 */
BOOL JavaMonkey::InitWindow(int windowMode) {
	theMonkeyWindow = CreateWindow(
		theMonkeyClassName,
		"Java Monkey",
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		HWND_DESKTOP,
		NULL,
		theInstance,
		(void *) NULL);

	// If window could not be created, return "failure".
	if (!theMonkeyWindow)
	   return (FALSE);

	/* Initialize the common controls. */
	INITCOMMONCONTROLSEX cc;
	cc.dwSize = sizeof(INITCOMMONCONTROLSEX);
	cc.dwICC = ICC_TREEVIEW_CLASSES;
	InitCommonControlsEx(&cc); 

	ShowWindow(theMonkeyWindow, windowMode);
	UpdateWindow(theMonkeyWindow); 

	return (TRUE);      
}


/**
 * Build a tree (and the treeview control) of all accessible Java components
 *
 */
void JavaMonkey::buildAccessibilityTree() {
	// have MS-Windows call EnumWndProc() with all of the top-level windows
	EnumWindows((WNDENUMPROC) EnumWndProc, NULL);
}

extern "C" {
	/**
	 * Create (and display) the accessible component nodes of a parent AccessibleContext
	 *
	 */
	BOOL CALLBACK EnumWndProc(HWND hWnd, LPARAM lParam) {
		if (IsJavaWindow(hWnd)) {
			long vmID;
			AccessibleContext ac;
			if (GetAccessibleContextFromHWND(hWnd, &vmID, &ac) == TRUE) {
				theMonkey->addComponentNodes(vmID, ac, (AccessibleNode *) NULL, 
											 hWnd, TVI_ROOT, theTreeWindowControl);
			}
		}
		return(TRUE);
	}
}


/**
 * Create (and display) the accessible component nodes of a parent AccessibleContext
 *
 */
void JavaMonkey::addComponentNodes(long vmID, AccessibleContext context,
								   AccessibleNode *parent, HWND hWnd, 
								   HTREEITEM treeNodeParent, HWND treeWnd) {
	AccessibleNode *newNode = new AccessibleNode(vmID, context, parent, hWnd, treeNodeParent);

	AccessibleContextInfo info;
	if (GetAccessibleContextInfo(vmID, context, &info) != FALSE) {
		char s[BUFSIZE];

        wsprintf(s, "%ls", info.name);
		newNode->setAccessibleName(s);
        wsprintf(s, "%ls", info.role);
		newNode->setAccessibleRole(s);

		wsprintf(s, "%ls [%ls]", info.name, info.role);

		TVITEM tvi;
		tvi.mask = TVIF_PARAM | TVIF_TEXT;	// text and lParam are only valid parts
		tvi.pszText = (char *) s;
		tvi.cchTextMax = strlen(s);
		tvi.lParam = (long) newNode;

		TVINSERTSTRUCT tvis;
		tvis.hParent = treeNodeParent;
		tvis.hInsertAfter = TVI_LAST;	// make tree in order given
		tvis.item = tvi;

		HTREEITEM treeNodeItem = TreeView_InsertItem(treeWnd, &tvis);

		for (int i = 0; i < info.childrenCount; i++) {
			addComponentNodes(vmID, GetAccessibleChildFromContext(vmID, context, i), 
							  newNode, hWnd, treeNodeItem, treeWnd);
		}
	} else {
		char s[BUFSIZE];
		sprintf(s, "ERROR calling GetAccessibleContextInfo; vmID = %X, context = %X", vmID, context);

		TVITEM tvi;
		tvi.mask = TVIF_PARAM | TVIF_TEXT;	// text and lParam are only valid parts
		tvi.pszText = (char *) s;
		tvi.cchTextMax = strlen(s);
		tvi.lParam = (long) newNode;

		TVINSERTSTRUCT tvis;
		tvis.hParent = treeNodeParent;
		tvis.hInsertAfter = TVI_LAST;	// make tree in order given
		tvis.item = tvi;

		HTREEITEM treeNodeItem = TreeView_InsertItem(treeWnd, &tvis);
	}
}

// -----------------------------

/**
 * Create an AccessibleNode
 *
 */
AccessibleNode::AccessibleNode(long JavaVMID, AccessibleContext context,
							   AccessibleNode *parent, HWND hWnd, 
							   HTREEITEM parentTreeNodeItem) {
	vmID = JavaVMID;
	ac = context;
	parentNode = parent;
	baseHWND = hWnd;
	treeNodeParent = parentTreeNodeItem;

	// setting accessibleName and accessibleRole not done here,
	// in order to minimize calls to the AccessBridge
	// (since such a call is needed to enumerate children)
}

/**
 * Destroy an AccessibleNode
 *
 */
AccessibleNode::~AccessibleNode() {
	ReleaseJavaObject(vmID, ac);
}

/**
 * Set the accessibleName string
 *
 */
void AccessibleNode::setAccessibleName(char *name) {
	strncpy(accessibleName, name, MAX_STRING_SIZE);
}

/**
 * Set the accessibleRole string
 *
 */
void AccessibleNode::setAccessibleRole(char *role) {
	strncpy(accessibleRole, role, SHORT_STRING_SIZE);
}

	
/**
 * Create an API window to show off the info for this AccessibleContext
 *
 */
BOOL AccessibleNode::displayAPIWindow() {
	HWND apiWindow;

	apiWindow = CreateWindow(
		theExplorerClassName,
		"Java Accessibility API view",
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		350,
		180,
		HWND_DESKTOP,
		NULL,
		theInstance,
		(void *) NULL);

	// If window could not be created, return "failure".
	if (!apiWindow)
	   return (FALSE);

	char buffer[BUFSIZE];
    buffer[0] = '\0';
	getAccessibleInfo(vmID, ac, buffer, BUFSIZE);
	SetDlgItemText(apiWindow, cExplorerText, buffer);														  

	ShowWindow(apiWindow, SW_SHOWNORMAL);
	UpdateWindow(apiWindow); 

	return (TRUE);      
}

// -----------------------------


/**
 * Gather the Accessible information about the object under the mouse
 */
char *getAccessibleInfo(long vmID, AccessibleContext ac, char *buffer, int bufsize) {
    char s[1024];
    AccessibleContext parentAC;
	AccessibleContextInfo info;
	AccessibleContextInfo parentInfo;
	AccessibleTextInfo textInfo;
	AccessibleTextItemsInfo textItems;
	AccessibleTextSelectionInfo textSelection;
	AccessibleTextRectInfo rectInfo;
    AccessibleTextAttributesInfo attributeInfo;
	wchar_t tempBuf[1024];

    if (ac == (AccessibleContext) 0) {
        strcpy(buffer, "(null)");
    } else {
		//
		// Interface Accessible - our stuff
		//
		if (GetAccessibleContextInfo(vmID, ac, &info) == FALSE) {
			strncat(buffer, "\r\nGetAccessibleContextInfo() failed.", bufsize);
		} else {
			strncat(buffer, "Accessible Context information:", bufsize);

				// Core info
            wsprintf(s, "\r\n  Name:  %ls", info.name);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n  Description:  %ls", info.description);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n  Role:  %ls", info.role);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n  Role in en_US locale:  %ls", info.role_en_US);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n  States:  %ls", info.states);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n  States in en_US locale:  %ls", info.states_en_US);
				strncat(buffer, s, bufsize);
			sprintf(s, "\r\n  Index in parent:  %d", info.indexInParent);
				strncat(buffer, s, bufsize);
			sprintf(s, "\r\n  Children Count:  %d", info.childrenCount);
				strncat(buffer, s, bufsize);
			sprintf(s, "\r\n  Bounding Rectangle:  [%d, %d, %d, %d]", 
				info.x, info.y, info.x + info.width, info.y + info.height);
				strncat(buffer, s, bufsize);

            // Get some info about the parent
            parentAC = GetAccessibleParentFromContext(vmID, ac);
            if (parentAC != 0) {
		        if (GetAccessibleContextInfo(vmID, parentAC, &parentInfo) == FALSE) {
			        strncat(buffer, "\r\n  GetAccessibleContextInfo() failed for parent.", bufsize);
		        } else {
			        sprintf(s, "\r\n  Parent's Name:  %ls", parentInfo.name);
				    strncat(buffer, s, bufsize);
/*
					// Simple test for IsSameObject
                    if (IsSameObject(vmID, ac, ac) == TRUE &&
						IsSameObject(vmID, ac, parentAC) == FALSE) {
						strncat(buffer, "\r\n  IsSameObject correct", bufsize);
					} else {
						strncat(buffer, "\r\n  IsSameObject incorrect", bufsize);
*/
					
                }
                ReleaseJavaObject(vmID, parentAC);
            } else {

			        strncat(buffer, "\r\n  GetAccessibleParentFromContext failed!!!", bufsize);
            }

				// Value info
			if (info.accessibleInterfaces | cAccessibleValueInterface) {

				strncat(buffer, "\r\nAccessible Value information:", bufsize);

				if (GetCurrentAccessibleValueFromContext(vmID, ac, tempBuf, sizeof(tempBuf)) == TRUE) {
					wsprintf(s, "\r\n  Current Value:  %ls", tempBuf);
				        strncat(buffer, s, bufsize);
				}
				if (GetMaximumAccessibleValueFromContext(vmID, ac, tempBuf, sizeof(tempBuf)) == TRUE) {
					wsprintf(s, "\r\n  Maximum Value:  %ls", tempBuf);
				        strncat(buffer, s, bufsize);
				}
				if (GetMinimumAccessibleValueFromContext(vmID, ac, tempBuf, sizeof(tempBuf)) == TRUE) {
					wsprintf(s, "\r\n  Minimum Value:  %ls", tempBuf);
				        strncat(buffer, s, bufsize);
				}
            }
/*
				// Selection info
			if (info.accessibleSelection == TRUE) {

				strncat(buffer, "\r\nAccessible Selection information:", bufsize);

				if ((selCount = GetAccessibleSelectionCountFromContext(vmID, ac)) != -1) {
					wsprintf(s, "\r\n  Selection count:  %d", selCount);
				        strncat(buffer, s, bufsize);

                    for (i = 0; i < selCount; i++) {
				        if ((selectedAC = GetAccessibleSelectionFromContext(vmID, ac, i)) != (jobject) 0) {
		                    if (GetAccessibleContextInfo(vmID, selectedAC, &info) == FALSE) {
                                wsprintf(s, "\r\n  GetAccessibleContextInfo() failed on selection #%d", i);
				                    strncat(buffer, s, bufsize);
		                    } else {
			                    wsprintf(s, "\r\n  Name of selection #%d: %ls", i, info.name);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n  Description of selection #%d: %ls", i, info.description);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n  Role of selection #%d: %ls", i, info.role);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n  States of selection #%d: %ls", i, info.states);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n  Index in parent of selection #%d: %d", i, info.indexInParent);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n  Children Count of selection #%d: %d", i, info.childrenCount);
				                    strncat(buffer, s, bufsize);
                            }
                            ReleaseJavaObject(vmID, selectedAC);
                        } else {
			                wsprintf(s, "\r\n  GetAccessibleSelectionFromContext() failed on selection #%d", i);
				                    strncat(buffer, s, bufsize);
                        }
				    }
                }
            }
*/
                // Text info
			if (info.accessibleText == TRUE) {
				
				strncat(buffer, "\r\nAccessible Text information:", bufsize);

				if (GetAccessibleTextInfo(vmID, ac, &textInfo, 0, 0) == TRUE) {

					wsprintf(s, "\r\n  Char count:  %d", textInfo.charCount);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n  Caret at index:  %d", textInfo.caretIndex);
				        strncat(buffer, s, bufsize);

				}
				
                strncat(buffer, "\r\n  Info at start of text:", bufsize);

                if (GetAccessibleTextItems(vmID, ac, &textItems, 0) == TRUE) {

					wsprintf(s, "\r\n    Letter:  %lc", textItems.letter);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n    Word:  %ls", textItems.word);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n    Sentence:  %ls", textItems.sentence);
				        strncat(buffer, s, bufsize);

				}

				if (GetAccessibleTextAttributes(vmID, ac, 
												textInfo.indexAtPoint,
												&attributeInfo)) {

					sprintf(s, "\r\n    Core attributes: %s", 
                        attributeInfo.bold == TRUE ? "bold" : "not bold");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.italic == TRUE ? "italic" : "not italic");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.underline == TRUE ? "underline" : "not underline");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.strikethrough == TRUE ? "strikethrough" : "not strikethrough");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.superscript == TRUE ? "superscript" : "not superscript");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.subscript == TRUE ? "subscript" : "not subscript");
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n    Background color:  %ls", attributeInfo.backgroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n    Foreground color:  %ls", attributeInfo.backgroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n    Font family:  %ls", attributeInfo.fontFamily);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Font size:  %d", attributeInfo.fontSize);
						strncat(buffer, s, bufsize);


                    sprintf(s, "\r\n    First line indent:  %f", attributeInfo.firstLineIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Left indent:  %f", attributeInfo.leftIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Right indent:  %f", attributeInfo.rightIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Line spacing:  %f", attributeInfo.lineSpacing);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Space above:  %f", attributeInfo.spaceAbove);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Space below:  %f", attributeInfo.spaceBelow);
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n    Full attribute string:  %ls", attributeInfo.fullAttributesString);
						strncat(buffer, s, bufsize);
				}

				strncat(buffer, "\r\n  Info at caret:", bufsize);

                if (GetAccessibleTextItems(vmID, ac, &textItems, textInfo.caretIndex) == TRUE) {

					wsprintf(s, "\r\n    Letter:  %lc", textItems.letter);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n    Word:  %ls", textItems.word);
				        strncat(s, s, bufsize);
					wsprintf(s, "\r\n    Sentence:  %ls", textItems.sentence);
				        strncat(buffer, s, bufsize);

				}

                if (GetAccessibleTextAttributes(vmID, ac, 
												textInfo.caretIndex,
                                                &attributeInfo)) {

					sprintf(s, "\r\n    Core attributes: %s", 
                        attributeInfo.bold == TRUE ? "bold" : "not bold");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.italic == TRUE ? "italic" : "not italic");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.underline == TRUE ? "underline" : "not underline");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.strikethrough == TRUE ? "strikethrough" : "not strikethrough");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.superscript == TRUE ? "superscript" : "not superscript");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.subscript == TRUE ? "subscript" : "not subscript");
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n    Background color:  %ls", attributeInfo.backgroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n    Foreground color:  %ls", attributeInfo.backgroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n    Font family:  %ls", attributeInfo.fontFamily);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Font size:  %d", attributeInfo.fontSize);
						strncat(buffer, s, bufsize);


                    sprintf(s, "\r\n    First line indent:  %f", attributeInfo.firstLineIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Left indent:  %f", attributeInfo.leftIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Right indent:  %f", attributeInfo.rightIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Line spacing:  %f", attributeInfo.lineSpacing);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Space above:  %f", attributeInfo.spaceAbove);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n    Space below:  %f", attributeInfo.spaceBelow);
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n    Full attribute string:  %ls", attributeInfo.fullAttributesString);
						strncat(buffer, s, bufsize);
				}

                if (GetAccessibleTextSelectionInfo(vmID, ac, &textSelection) == TRUE) {

					wsprintf(s, "\r\n  Selection start index:  %d", textSelection.selectionStartIndex);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n  Selection end index:  %d", textSelection.selectionEndIndex);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n  Selected text:  %ls", textSelection.selectedText);
				        strncat(buffer, s, bufsize);

				}

				if (GetAccessibleTextRect(vmID, ac, &rectInfo, textInfo.indexAtPoint) == TRUE) {

					wsprintf(s, "\r\n  Bounding rect at index:  [%d, %d, %d, %d]",
						rectInfo.x, rectInfo.y, rectInfo.x + rectInfo.width, 
						rectInfo.y + rectInfo.height);
				        strncat(buffer, s, bufsize);

				}
            }

			// AccessibleTable
			AccessibleTableInfo tableInfo;
			if (info.accessibleInterfaces & cAccessibleTableInterface) {
				if (getAccessibleTableInfo(vmID, ac, &tableInfo) == FALSE) {

					strncat(buffer, "\r\ngetAccessibleTableInfo() failed.", bufsize);
				} else {
					strncat(buffer, "\r\nAccessibleTable info:", bufsize);

					// Core info
				 	wsprintf(s, "\r\n  #table rows:  %d", tableInfo.rowCount);
				 	strncat(buffer, s, bufsize);
				 	wsprintf(s, "\r\n  #table cols:  %d", tableInfo.columnCount);
					strncat(buffer, s, bufsize);

					AccessibleTableCellInfo tableCellInfo;
					for (int i = 0; i < tableInfo.rowCount; i++) {
						for (int j = 0; j < tableInfo.columnCount; j++) {

							if (getAccessibleTableCellInfo(vmID, tableInfo.accessibleTable, i, j, 
								&tableCellInfo) == FALSE) {

								strncat(buffer, "\r\nGetAccessibleTableCellInfo() failed.", bufsize);
							} else {
	
								wsprintf(s, "\r\n\r\n  AccessibleTable cell[%d,%d] info:", i, j);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  index: %ld", tableCellInfo.index);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  rowExtent: %ld", tableCellInfo.rowExtent);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  colExtent: %ld", tableCellInfo.columnExtent);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  isSel: %ld", tableCellInfo.isSelected);
								strncat(buffer, s, bufsize);
		
								AccessibleContextInfo cellACInfo;
					  			if (GetAccessibleContextInfo(vmID, tableCellInfo.accessibleContext, &cellACInfo) == FALSE) {
									strncat(buffer, "\r\nGetAccessibleContextInfo() failed.", bufsize);
								} else { 
									wsprintf(s, "\r\n  Name:  %ls", cellACInfo.name);
									strncat(buffer, s, bufsize);
									wsprintf(s, "\r\n  Description:  %ls", cellACInfo.description);
									strncat(buffer, s, bufsize);
									wsprintf(s, "\r\n  Role:  %ls", cellACInfo.role);
									strncat(buffer, s, bufsize);
									wsprintf(s, "\r\n  Role in en_US locale:  %ls", cellACInfo.role_en_US);
									strncat(buffer, s, bufsize);
									wsprintf(s, "\r\n  States:  %ls", cellACInfo.states);
									strncat(buffer, s, bufsize);
									wsprintf(s, "\r\n  States in en_US locale:  %ls", cellACInfo.states_en_US);
									strncat(buffer, s, bufsize);
									sprintf(s, "\r\n  Index in parent:  %d", cellACInfo.indexInParent);
									strncat(buffer, s, bufsize);
									sprintf(s, "\r\n  Children Count:  %d", cellACInfo.childrenCount);
									strncat(buffer, s, bufsize);
									sprintf(s, "\r\n  Bounding Rectangle:  [%d, %d, %d, %d]", 
									    cellACInfo.x, cellACInfo.y, cellACInfo.x + cellACInfo.width, cellACInfo.y + cellACInfo.height);
									strncat(buffer, s, bufsize);
				
								}
							}
						}
					}

					// Get the column headers
					AccessibleTableInfo columnInfo;
					if (getAccessibleTableColumnHeader(vmID, ac, &columnInfo) == FALSE) {
						strncat(buffer, "\r\ngetAccessibleTableColumnHeader() failed.", bufsize);
					} else {
						strncat(buffer, "\r\n\r\nAccessibleTable column header info:", bufsize);

						// Core info
				 		wsprintf(s, "\r\n  #column header rows:  %d", columnInfo.rowCount);
				 		strncat(buffer, s, bufsize);
				 		wsprintf(s, "\r\n  #column header cols:  %d", columnInfo.columnCount);
						strncat(buffer, s, bufsize);

					}

					// Get the selected rows
					int numSelections = getAccessibleTableRowSelectionCount(vmID, 
						tableInfo.accessibleTable);
					wsprintf(s, "\r\n\r\n#row selections:  %d", numSelections);
					strncat(buffer, s, bufsize);
					jint *selections = new jint[numSelections];

					if (getAccessibleTableRowSelections(vmID, tableInfo.accessibleTable, 
						numSelections, selections) == FALSE) {
						strncat(buffer, "\r\ngetAccessibleTableRowSelections() failed.", 
							bufsize);
					} else {
						wsprintf(s, "  \r\n  row selections: ");
						strncat(buffer, s, bufsize);
						for (int j = 0; j < numSelections; j++) {
							wsprintf(s, " %d", selections[j]);
							strncat(buffer, s, bufsize);
						}
					}

					// Get column header info
					for (i = 0; i < columnInfo.columnCount; i++) {
						if (getAccessibleTableCellInfo(vmID, columnInfo.accessibleTable, 0, i, 
							&tableCellInfo) == FALSE) {

							strncat(buffer, "\r\nGetAccessibleTableCellInfo() failed.", bufsize);
						} else {
							wsprintf(s, "\r\n\r\nColumn header [0,%d] cell info.", i);
							strncat(buffer, s, bufsize);
							wsprintf(s, "\r\n  index: %ld", tableCellInfo.index);
							strncat(buffer, s, bufsize);
							wsprintf(s, "\r\n  rowExtent: %ld", tableCellInfo.rowExtent);
							strncat(buffer, s, bufsize);
							wsprintf(s, "\r\n  colExtent: %ld", tableCellInfo.columnExtent);
							strncat(buffer, s, bufsize);
							wsprintf(s, "\r\n  isSel: %ld", tableCellInfo.isSelected);
							strncat(buffer, s, bufsize);

							AccessibleContextInfo cellACInfo;
				  			if (GetAccessibleContextInfo(vmID, tableCellInfo.accessibleContext, &cellACInfo) == FALSE) {
								strncat(buffer, "\r\nGetAccessibleContextInfo() failed.", bufsize);
							} else { 
								wsprintf(s, "\r\n  Name:  %ls", cellACInfo.name);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  Description:  %ls", cellACInfo.description);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  Role:  %ls", cellACInfo.role);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  Role in en_US locale:  %ls", cellACInfo.role_en_US);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  States:  %ls", cellACInfo.states);
								strncat(buffer, s, bufsize);
								wsprintf(s, "\r\n  States in en_US locale:  %ls", cellACInfo.states_en_US);
								strncat(buffer, s, bufsize);
								sprintf(s, "\r\n  Index in parent:  %d", cellACInfo.indexInParent);
								strncat(buffer, s, bufsize);
								sprintf(s, "\r\n  Children Count:  %d", cellACInfo.childrenCount);
								strncat(buffer, s, bufsize);
								sprintf(s, "\r\n  Bounding Rectangle:  [%d, %d, %d, %d]", 
						            cellACInfo.x, cellACInfo.y, cellACInfo.x + cellACInfo.width, cellACInfo.y + cellACInfo.height);
								strncat(buffer, s, bufsize);
	
							}
						}
					}
				}
			}
		}
	}

	return buffer;													  
}

// -----------------------------

void JavaMonkey::exitJavaMonkey(HWND hWnd) {
	EndDialog(hWnd, TRUE);
	PostQuitMessage (0);
}

// -----------------------------

extern "C" {

	// CreateATreeView - creates a tree-view control. 
	// Returns the handle of the new control if successful or NULL//     otherwise. 
	// hwndParent - handle of the control's parent window 
	HWND CreateATreeView(HWND hwndParent) { 
		RECT rcClient;  // dimensions of client area 
		HWND hwndTV;    // handle of tree-view control 
 
		// Get the dimensions of the parent window's client area, and create 
		// the tree-view control. 
		GetClientRect(hwndParent, &rcClient); 
		hwndTV = CreateWindow(WC_TREEVIEW, 
							  "",
							  WS_VISIBLE | WS_TABSTOP | WS_CHILD |
							  TVS_HASLINES | TVS_HASBUTTONS |
							  TVS_LINESATROOT, 
							  0, 0, rcClient.right, rcClient.bottom, 
							  hwndParent,
							  (HMENU) cTreeControl,
							  theInstance,
							  NULL);

		return hwndTV; 
	} 


	LRESULT CALLBACK MonkeyWindowProc(HWND hWnd, UINT message, 
									  UINT wParam, LONG lParam) {
		int command;
		short width, height;

		switch (message) {
		case WM_CREATE:
			theTreeWindowControl = CreateATreeView(hWnd);
			break;

		case WM_CLOSE:
			theMonkey->exitJavaMonkey(hWnd);
			break;
		
		case WM_SIZE:
			width = LOWORD(lParam);
			height = HIWORD(lParam);
			SetWindowPos(theTreeWindowControl, NULL, 0, 0, width, height, 0); 
			return(FALSE);			// let windows finish handling this
			break;
		
		case WM_COMMAND:
			command = LOWORD(wParam);
			switch(command) {

			case cExitMenuItem:
				theMonkey->exitJavaMonkey(hWnd);
				break;

			case cRefreshTreeItem:
				// update the tree
				theMonkey->buildAccessibilityTree();
				break;

			case cAccessibilityAPItem:
				// open a new window with the Accessibility API in it for the
				// selected element in the tree
				if (theSelectedNode != (AccessibleNode *) 0) {
					theSelectedNode->displayAPIWindow();
				}
				break;
			
			}
			break;

		case WM_NOTIFY:				// receive tree messages
			NMTREEVIEW *nmptr;
			nmptr = (LPNMTREEVIEW) lParam;
			if (nmptr->hdr.code == TVN_SELCHANGED) {
				theSelectedNode = (AccessibleNode *) nmptr->itemNew.lParam;	
			}
			break;

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}

		return 0;
	}


	LRESULT CALLBACK ExplorerWindowProc(HWND hWnd, UINT message, 
										UINT wParam, LONG lParam) {
		short width, height;
		HWND dlgItem;

		switch (message) {
		case WM_CREATE:
			RECT rcClient;    // dimensions of client area 
			HWND hwndEdit;    // handle of tree-view control 
 
			// Get the dimensions of the parent window's client area,  
			// and create the edit control. 
			GetClientRect(hWnd, &rcClient); 
			hwndEdit = CreateWindow("Edit", 
								    "",
								    WS_VISIBLE | WS_TABSTOP | WS_CHILD |
								    ES_MULTILINE | ES_AUTOVSCROLL | 
								    ES_READONLY | WS_VSCROLL,
								    0, 0, rcClient.right, rcClient.bottom, 
								    hWnd,
								    (HMENU) cExplorerText,
								    theInstance,
								    NULL);
			break;

		case WM_CLOSE:
			DestroyWindow(hWnd);
			break;
		
		case WM_SIZE:
			width = LOWORD(lParam);
			height = HIWORD(lParam);
			dlgItem = GetDlgItem(hWnd, cExplorerText); 
			SetWindowPos(dlgItem, NULL, 0, 0, width, height, 0); 
			return(FALSE);			// let windows finish handling this
			break;

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}

		return 0;   
	}

}

// -----------------------------

extern "C" {

	/**
	 * WinMain
	 *
	 */
	int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, 
					   LPSTR lpCmdLine, int nShowCmd) {
		MSG msg;

		theInstance = hInst;
		theMonkey = new JavaMonkey(hInst);
		if (theMonkey->InitWindow(nShowCmd) == FALSE) {
			return(0);
		}
		if (initializeAccessBridge() == TRUE) {
			while (GetMessage(&msg, NULL, 0, 0)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
			shutdownAccessBridge();
		}
		return(0);
	}
}

